ATTACHMENTS = "_attachments"
DURATION = "_duration"
EDITED = "_edited"
END = "end"
END_TIME = "end_time"
GEOLOCATION = "_geolocation"
GROUP_DELIMETER_TAG = 'group_delimiter'
ID = "_id"
KNOWN_MEDIA_TYPES = ['photo', 'image', 'audio', 'video']
LAST_EDITED = "_last_edited"
MEDIA_ALL_RECEIVED = "_media_all_received"
MEDIA_COUNT = "_media_count"
NOTES = "_notes"
REPEAT_INDEX_TAGS = 'repeat_index_tags'
START = "start"  # This tag was used in Phase II
START_TIME = "start_time"  # This tag was used in Phase I
STATUS = "_status"
SUBMISSION_TIME = '_submission_time'
SUBMITTED_BY = "_submitted_by"
TOTAL_MEDIA = "_total_media"
UUID = "_uuid"
VERSION = "_version"
XFORM_ID = "_xform_id"
XFORM_ID_STRING = "_xform_id_string"
