# Generated by Django 2.2 on 2019-04-29 19:18

from django.db import migrations, models
import django.db.models.deletion
import xform.models


class Migration(migrations.Migration):

    dependencies = [
        ('xform', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Attachment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('media_file', models.FileField(max_length=255, upload_to=xform.models.upload_to)),
                ('mimetype', models.CharField(blank=True, default='', max_length=100)),
                ('extension', models.CharField(db_index=True, default='non', max_length=10)),
                ('date_created', models.DateTimeField(auto_now_add=True, null=True)),
                ('date_modified', models.DateTimeField(auto_now=True, null=True)),
                ('file_size', models.PositiveIntegerField(default=0)),
                ('name', models.CharField(blank=True, max_length=100, null=True)),
                ('instance', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='attachments', to='xform.Instance')),
            ],
            options={
                'ordering': ('pk',),
            },
        ),
    ]
