# Generated by Django 2.2 on 2019-05-06 23:01

from django.db import migrations, models
import xform.models


class Migration(migrations.Migration):

    dependencies = [
        ('xform', '0003_datadictionary'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metadata',
            name='data_file',
            field=models.FileField(blank=True, null=True, upload_to=xform.models.meta_data_upload_to),
        ),
    ]
